import math
from typing import Callable, List, Dict, Iterable

import torch

from .GenericPSO import clone_param_groups, GenericPSO
from .SineCosineAlgorithm import SCAParticle


class ImprovedSCAParticle(SCAParticle):
    def __init__(self,
                 param_groups,
                 max_param_value: float,
                 min_param_value: float):

        super().__init__(param_groups, max_param_value, min_param_value)

    def step(self, closure: Callable[[], torch.Tensor], global_best_param_groups: List[Dict],
             r1: torch.Tensor, r2: torch.Tensor, r3: torch.Tensor, use_sine: bool, w: float) -> torch.Tensor:
        """
        Particle will take one step.
        :param closure: A callable that reevaluates the model and returns the loss.
        :param global_best_param_groups: List of param_groups that yield the best found loss globally
        :return:
        """
        # Because our parameters are not a single tensor, we have to iterate over each group, and then each param in
        # each group.
        for position_group, global_best, master in zip(self.position,
                                                       global_best_param_groups,
                                                       self.param_groups):
            position_group_params = position_group['params']
            global_best_params = global_best['params']
            master_params = master['params']

            new_position_params = []
            for p, gb, m in zip(position_group_params,
                                global_best_params, master_params):
                func = torch.sin if use_sine else torch.cos
                new_position = w * p + r1 * func(r2) * abs(r3 * gb - p)
                new_position_params.append(new_position)
                m.data = new_position.data  # Update the model, so we can use it for calculating loss
            position_group['params'] = new_position_params

        # Really crummy way to update the parameter weights in the original model.
        # Simply changing self.param_groups doesn't update the model.
        # Nor does changing its elements or the raw values of 'param' of the elements.
        # We have to change the underlying tensor data to point to the new positions
        for i in range(len(self.position)):
            for j in range(len(self.param_groups[i]['params'])):
                self.param_groups[i]['params'][j].data = self.param_groups[i]['params'][j].data

        # Calculate new loss after moving and update the best known position if we're in a better spot
        new_loss = closure()
        if new_loss < self.best_known_loss_value:
            self.best_known_position = clone_param_groups(self.position)
            self.best_known_loss_value = new_loss
        return new_loss


class ImprovedSineCosineAlgorithm(GenericPSO):
    """
    The Improved Sine Cosine Algorithm is a variation on the original Sine Cosine Algorithm, specifically attempting to
    improve its resilience against local optima and unbalanced exploitation. It also demonstrates better behavior in
    high-dimensional optimization problems.

    Original Paper:
    Long, W., Wu, T., Liang, X., & Xu, S. (2019). Solving high-dimensional global optimization problems using an
    improved sine cosine algorithm. Expert systems with applications, 123, 108-126.

    https://e-tarjome.com/storage/panel/fileuploads/2019-08-22/1566462251_E11587-e-tarjome.pdf
    """

    def __init__(self,
                 params: Iterable[torch.nn.Parameter],
                 num_particles: int = 100,
                 a_start: float = 2,
                 a_end: float = 0,
                 max_time_steps: int = 1000,
                 w_end: float = 0,
                 w_start: float = .1,
                 k: float = 15.,
                 max_param_value: float = -10,
                 min_param_value: float = 10):
        particle_kwargs = {
            'max_param_value': max_param_value,
            'min_param_value': min_param_value,
        }
        super().__init__(params, num_particles, particle_class=ImprovedSCAParticle, particle_kwargs=particle_kwargs)
        self.max_param_value = max_param_value
        self.min_param_value = min_param_value
        self.magnitude = max_param_value - min_param_value
        self.w_end = w_end
        self.w_start = w_start
        self.a_start = a_start
        self.a_end = a_end
        self.k = k

        self.max_time_steps = max_time_steps
        self.current_time_step = 0

    def w(self) -> float:
        """
        Calculate the inertial weight of the particle at the current time step.
        """
        return self.w_end + (self.w_start - self.w_end) * (
                    self.max_time_steps - self.current_time_step) / self.max_time_steps

    def r1(self):
        """Calculate the r1 value for the current timestep"""
        return (self.a_start - self.a_end) * math.exp(
            -self.current_time_step ** 2 / (self.k * self.max_time_steps) ** 2) + self.a_end

    @torch.no_grad()
    def step(self, closure: Callable[[], torch.Tensor]) -> torch.Tensor:
        """
        Performs a single optimization step.

        :param closure: A callable that reevaluates the model and returns the loss.
        :return: the final loss after the step (as calculated by the closure)
        """
        r3 = 2 * torch.rand((1,))
        r2 = 2 * torch.pi * torch.rand((1,))
        use_sine = torch.rand((1,)).item() < .5
        r1 = self.r1()
        w = self.w()
        self.current_time_step += 1
        return super().step(closure, particle_step_kwargs={'r1': r1,
                                                           'r2': r2,
                                                           'r3': r3,
                                                           'use_sine': use_sine,
                                                           'w': w
                                                           })



from abc import ABC, abstractmethod, ABCMeta
from typing import Dict, List, Callable, Type, Iterable, Optional

import torch
from torch.optim import Optimizer


def clone_param_group(param_group: Dict) -> Dict:
    """
    Clone each param in a param_group and return a new dict containing the clones
    :param param_group: Dict containing param_groups
    :return: cloned param_group dict
    """
    new_group = {key: value for key, value in param_group.items() if key != 'params'}
    new_group['params'] = [param.detach().clone() for param in param_group['params']]
    return new_group


def clone_param_groups(param_groups: List[Dict]) -> List[Dict]:
    """
    Make a list of clones for each param_group in a param_groups list.
    :param param_groups: List of dicts containing param_groups
    :return: cloned list of param_groups
    """
    return [clone_param_group(param_group) for param_group in param_groups]


def _initialize_param_groups(param_groups: List[Dict], max_param_value, min_param_value) -> List[Dict]:
    """
    Take a list of param_groups, clone it, and then randomly initialize its parameters with values between
    max_param_value and min_param_value.

    :param param_groups: List of dicts containing param_groups
    :param max_param_value: Maximum value of the parameters in the search space
    :param min_param_value: Minimum value of the parameters in the search space
    :return the new, initialized param_groups
    """
    magnitude = max_param_value - min_param_value
    mean_value = (max_param_value + min_param_value) / 2

    def _initialize_params(param):
        return magnitude * torch.rand_like(param) - magnitude / 2 + mean_value

    # Make sure we get a clone, so we don't overwrite the original params in the module
    param_groups = clone_param_groups(param_groups)
    for group in param_groups:
        group['params'] = [_initialize_params(p) for p in group['params']]

    return param_groups


class GenericParticle(ABC):
    def __init__(self, *args, **kwargs):
        self.position: List[Dict] = []

    @abstractmethod
    def step(self, closure: Callable[[], torch.Tensor], global_best_param_groups: List[Dict], **kwargs) -> torch.Tensor:
        """
        Particle will take one step.
        :param closure: A callable that reevaluates the model and returns the loss.
        :param global_best_param_groups: List of param_groups that yield the best found loss globally
        :return:
        """
        pass


class GenericPSO(Optimizer):
    def __init__(self, params: Iterable[torch.nn.Parameter], num_particles: int, particle_class: Type[GenericParticle],
                 particle_args: Optional[List] = None, particle_kwargs: Optional[Dict] = None):
        defaults = {}
        super().__init__(params, defaults)
        if particle_args is None:
            particle_args = []
        if particle_kwargs is None:
            particle_kwargs = {}
        self.particles = [particle_class(self.param_groups, *particle_args, **particle_kwargs)
                          for _ in range(num_particles)]

        self.best_known_global_param_groups = clone_param_groups(self.param_groups)
        self.best_known_global_loss_value = torch.inf

    @torch.no_grad()
    def step(self, closure: Callable[[], torch.Tensor], particle_step_kwargs: Optional[Dict] = None) -> torch.Tensor:
        """
        Performs a single optimization step.

        :param particle_step_kwargs: Dict of keyword arguments to pass to the particle step function, if needed.
        :param closure: A callable that reevaluates the model and returns the loss.
        :return: the final loss after the step (as calculated by the closure)
        """
        if particle_step_kwargs is None:
            particle_step_kwargs = {}
        for particle in self.particles:
            particle_loss = particle.step(closure, self.best_known_global_param_groups, **particle_step_kwargs)
            if particle_loss < self.best_known_global_loss_value:
                self.best_known_global_param_groups = clone_param_groups(particle.position)
                self.best_known_global_loss_value = particle_loss

        # set the module's parameters to be the best performing ones
        for master_group, best_group in zip(self.param_groups, self.best_known_global_param_groups):
            clone = clone_param_group(best_group)['params']
            for i in range(len(clone)):
                master_group['params'][i].data = clone[i].data

        return closure()  # loss = closure()

    subclasses = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses.append(cls)



from typing import Callable, Iterable, List, Dict

import torch

from .ParticleSwarmOptimizer import ParticleSwarmOptimizer
from .GenericPSO import clone_param_group, clone_param_groups


class RingTopologyPSO(ParticleSwarmOptimizer):
    """
    Ring Topology PSO is a modification of the naive Particle Swarm Optimization Algorithm, where instead of feeding
    each particle the global optimum, each particle only receives the minimum of itself and two other particles. These
    two particles are the same through the entire iteration.

    This is called a "ring" because the network graph showing these connections has a single cycle.
    """

    def __init__(self,
                 params: Iterable[torch.nn.Parameter],
                 num_neighbors: int = 2,
                 inertial_weight: float = .9,
                 cognitive_coefficient: float = 1.,
                 social_coefficient: float = 1.,
                 num_particles: int = 100,
                 max_param_value: float = 10.,
                 min_param_value: float = -10.):
        super().__init__(params, inertial_weight, cognitive_coefficient, social_coefficient, num_particles,
                         max_param_value, min_param_value)
        self.losses = {i: (particle.position, torch.inf) for i, particle in enumerate(self.particles)}
        self.num_neighbors = num_neighbors

    def _find_minimum_of_neighbors(self, particle_index: int) -> List[Dict]:
        neighbors = [(particle_index + i) % len(self.particles) for i in range(self.num_neighbors)]

        best = sorted([self.losses[n] for n in neighbors],
                      key=lambda x: x[1],
                      reverse=True)[0]
        return clone_param_groups(best[0])

    @torch.no_grad()
    def step(self, closure: Callable[[], torch.Tensor]) -> torch.Tensor:
        """
        Performs a single optimization step.

        :param closure: A callable that reevaluates the model and returns the loss.
        :return: the final loss after the step (as calculated by the closure)
        """
        losses = {}
        for i, particle in enumerate(self.particles):
            particle_loss = particle.step(closure, self._find_minimum_of_neighbors(i))
            losses[i] = (particle.position, particle_loss)
            if particle_loss < self.best_known_global_loss_value:
                self.best_known_global_param_groups = clone_param_groups(particle.position)
                self.best_known_global_loss_value = particle_loss

        # set the module's parameters to be the best performing ones
        for master_group, best_group in zip(self.param_groups, self.best_known_global_param_groups):
            clone = clone_param_group(best_group)['params']
            for i in range(len(clone)):
                master_group['params'][i].data = clone[i].data
        self.losses = losses
        return closure()  # loss = closure()



from typing import Callable, List, Dict, Iterable

import torch

from .GenericPSO import clone_param_groups, _initialize_param_groups, GenericParticle, GenericPSO


class SCAParticle(GenericParticle):
    def __init__(self,
                 param_groups,
                 max_param_value: float,
                 min_param_value: float):

        self.param_groups = param_groups
        self.max_param_value = max_param_value
        self.min_param_value = min_param_value

        self.position = _initialize_param_groups(param_groups, max_param_value, min_param_value)

        self.best_known_position = clone_param_groups(self.position)
        self.best_known_loss_value = torch.inf

    def step(self, closure: Callable[[], torch.Tensor], global_best_param_groups: List[Dict],
             r1: torch.Tensor, r2: torch.Tensor, r3: torch.Tensor, use_sine: bool) -> torch.Tensor:
        """
        Particle will take one step.
        :param closure: A callable that reevaluates the model and returns the loss.
        :param global_best_param_groups: List of param_groups that yield the best found loss globally
        :return:
        """
        # Because our parameters are not a single tensor, we have to iterate over each group, and then each param in
        # each group.
        for position_group, global_best, master in zip(self.position,
                                                       global_best_param_groups,
                                                       self.param_groups):
            position_group_params = position_group['params']
            global_best_params = global_best['params']
            master_params = master['params']

            new_position_params = []
            for p, gb, m in zip(position_group_params,
                                global_best_params, master_params):
                func = torch.sin if use_sine else torch.cos
                new_position = p + r1 * func(r2) * abs(r3 * gb - p)
                new_position_params.append(new_position)
                m.data = new_position.data  # Update the model, so we can use it for calculating loss
            position_group['params'] = new_position_params

        # Really crummy way to update the parameter weights in the original model.
        # Simply changing self.param_groups doesn't update the model.
        # Nor does changing its elements or the raw values of 'param' of the elements.
        # We have to change the underlying tensor data to point to the new positions
        for i in range(len(self.position)):
            for j in range(len(self.param_groups[i]['params'])):
                self.param_groups[i]['params'][j].data = self.param_groups[i]['params'][j].data

        # Calculate new loss after moving and update the best known position if we're in a better spot
        new_loss = closure()
        if new_loss < self.best_known_loss_value:
            self.best_known_position = clone_param_groups(self.position)
            self.best_known_loss_value = new_loss
        return new_loss


class SineCosineAlgorithm(GenericPSO):
    """
    The Sine Cosine Algorithm is an algorithm, conceived by Seyedali Mirjalili, similar to PSO, although the author
    does not use particle terminology in the paper. Each particle moves towards a destination (the best candidate
    solution) in a step size determined by several random parameters. The "sine" and "cosine" in the name come from the
    use of sine and cosine functions to determine whether the particle's step will explore or exploit the currently
    best solution.

    Original Paper:
    Seyedali Mirjalili , SCA: A Sine Cosine Algorithm for Solving Optimization Problems, Knowledge-Based Systems (2016),
    doi: 10.1016/j.knosys.2015.12.022

    https://dl.programstore.ir/files/Uploades/Lib/PDF/SCA.pdf
    """

    def __init__(self,
                 params: Iterable[torch.nn.Parameter],
                 num_particles: int = 100,
                 max_movement_radius: float = 2,
                 max_param_value: float = -10,
                 min_param_value: float = 10):
        particle_kwargs = {
            'max_param_value': max_param_value,
            'min_param_value': min_param_value,
        }
        super().__init__(params, num_particles, particle_class=SCAParticle, particle_kwargs=particle_kwargs)
        self.max_param_value = max_param_value
        self.min_param_value = min_param_value
        self.max_movement_radius = max_movement_radius
        self.magnitude = max_param_value-min_param_value
        self.initial_movement_radius = max_movement_radius


    @torch.no_grad()
    def step(self, closure: Callable[[], torch.Tensor]) -> torch.Tensor:
        """
        Performs a single optimization step.

        :param closure: A callable that reevaluates the model and returns the loss.
        :return: the final loss after the step (as calculated by the closure)
        """
        r3 = 2 * torch.rand((1,))
        r2 = 2*torch.pi*torch.rand((1,))
        use_sine = torch.rand((1,)).item() < .5
        r1 = self.max_movement_radius*torch.rand((1,))
        # self.max_movement_radius *= .99
        return super().step(closure, particle_step_kwargs={'r1': r1, 'r2': r2, 'r3': r3, 'use_sine': use_sine})



from typing import Callable, List, Dict, Iterable

import torch

from .GenericPSO import clone_param_groups, _initialize_param_groups, GenericParticle, GenericPSO


class ChaoticParticle(GenericParticle):
    def __init__(self,
                 param_groups,
                 a: float,
                 b: float,
                 c: float,
                 beta: float,
                 k: float,
                 epsilon: float,
                 i0: float,
                 z: float,
                 max_param_value: float,
                 min_param_value: float):
        if a <= 0:
            raise ValueError(f'A must be a positive constant, not of value {a}.')
        if b <= 0:
            raise ValueError(f'B must be a positive constant, not of value {b}.')
        if c <= 0:
            raise ValueError(f'C must be a positive constant, not of value {c}.')
        if not 0 < beta < 1:
            raise ValueError(f'Beta must be a positive constant, not of value {beta}.')

        self.param_groups = param_groups
        self.a = a
        self.b = b
        self.c = c
        self.beta = beta
        self.k = k
        self.i0 = i0
        self.epsilon = epsilon
        self._z = z
        self._z0 = z
        self.max_param_value = max_param_value
        self.min_param_value = min_param_value

        magnitude = abs(max_param_value - min_param_value)
        self.position = _initialize_param_groups(param_groups, max_param_value, min_param_value)
        # self.velocity = _initialize_param_groups(param_groups, magnitude, -magnitude)
        self.velocity = _initialize_param_groups(param_groups, 0, 0)
        # I'm not sure what x_ip and u_ip represent, but their iterations are defined mathematically
        # Their initial values are not defined, however
        self.x_ip = _initialize_param_groups(param_groups, max_param_value, min_param_value)
        self.u_ip = _initialize_param_groups(param_groups, 0, 0)

        self.best_known_position = clone_param_groups(self.position)
        self.best_known_loss_value = torch.inf

    def _unit_interval_to_min_max(self, x):
        return self.min_param_value + x * (self.max_param_value - self.min_param_value)

    def _min_max_to_unit_interval(self, x):
        return (x - self.min_param_value) / (self.max_param_value - self.min_param_value)

    def step(self, closure: Callable[[], torch.Tensor], global_best_param_groups: List[Dict]) -> torch.Tensor:
        """
        Particle will take one step.
        :param closure: A callable that reevaluates the model and returns the loss.
        :param global_best_param_groups: List of param_groups that yield the best found loss globally
        :return:
        """
        # Because our parameters are not a single tensor, we have to iterate over each group, and then each param in
        # each group.
        for (position_group,
             velocity_group,
             x_ip_group,
             u_ip_group,
             personal_best,
             global_best,
             master) in zip(self.position,
                            self.velocity,
                            self.u_ip,
                            self.x_ip,
                            self.best_known_position,
                            global_best_param_groups,
                            self.param_groups):

            position_group_params = position_group['params']
            velocity_group_params = velocity_group['params']
            x_ip_group_params = x_ip_group['params']
            u_ip_group_params = u_ip_group['params']
            personal_best_params = personal_best['params']
            global_best_params = global_best['params']
            master_params = master['params']

            new_position_params = []
            new_velocity_params = []
            new_x_ip_params = []
            new_u_ip_params = []
            for x, u, x_ip, u_ip, pb, gb, m in zip(position_group_params,
                                                   velocity_group_params,
                                                   x_ip_group_params,
                                                   u_ip_group_params,
                                                   personal_best_params,
                                                   global_best_params,
                                                   master_params):
                # rand_personal = torch.rand_like(u)
                # rand_group = torch.rand_like(u)
                # new_velocity = (self.inertial_weight * u
                #                 + self.cognitive_coefficient * rand_personal * (pb - x)
                #                 + self.social_coefficient * rand_group * (gb - x)
                #                 )
                # new_position = x + new_velocity
                # All the below calculations assume x is between 0 and 1
                x = self._min_max_to_unit_interval(x)
                x_ip = self._min_max_to_unit_interval(x_ip)
                gb = self._min_max_to_unit_interval(gb)
                pb = self._min_max_to_unit_interval(pb)
                # u = self._min_max_to_unit_interval(u)
                # u_ip = self._min_max_to_unit_interval(u_ip)

                delta_u = -(2 * self.a * (x - gb) + 2 * self.c * (x - x_ip)) / (1 + self.k * (self.a + self.b))
                new_velocity = u + delta_u - self._z * (x - self.i0)
                new_position = torch.clamp(self.k * new_velocity, min=0, max=1)
                new_position = self._unit_interval_to_min_max(new_position)
                # new_velocity = self._unit_interval_to_min_max(new_velocity)

                delta_u_ip = -(2 * self.b * (x - pb) + 2 * self.c * (x - x_ip)) / (1 + self.k * (self.b + self.c))
                new_u_ip = u_ip + delta_u_ip - self._z * (x_ip - self.i0)
                new_x_ip = torch.clamp(self.k * new_u_ip, min=0, max=1)
                new_x_ip = self._unit_interval_to_min_max(new_x_ip)
                # new_u_ip = self._unit_interval_to_min_max(new_u_ip)

                self._z *= 1 - self.beta

                self.k = 30 - 15*(self._z0 - self._z)/self._z0

                new_velocity_params.append(new_velocity)
                new_position_params.append(new_position)
                new_x_ip_params.append(new_x_ip)
                new_u_ip_params.append(new_u_ip)
                m.data = new_position.data  # Update the model, so we can use it for calculating loss
            position_group['params'] = new_position_params
            velocity_group['params'] = new_velocity_params
            x_ip_group['params'] = new_x_ip_params
            u_ip_group['params'] = new_u_ip_params

        # Really crummy way to update the parameter weights in the original model.
        # Simply changing self.param_groups doesn't update the model.
        # Nor does changing its elements or the raw values of 'param' of the elements.
        # We have to change the underlying tensor data to point to the new positions
        for i in range(len(self.position)):
            for j in range(len(self.param_groups[i]['params'])):
                self.param_groups[i]['params'][j].data = self.param_groups[i]['params'][j].data

        # Calculate new loss after moving and update the best known position if we're in a better spot
        new_loss = closure()
        if new_loss < self.best_known_loss_value:
            self.best_known_position = clone_param_groups(self.position)
            self.best_known_loss_value = new_loss
        return new_loss


class ChaoticPSO(GenericPSO):
    """
    ChaoticPSO is a variation on the original Particle Swarm Optimization inspired by techniques in training
    Hopfield Networks. It introduces some chaos-like mechanics into the optimization, theoretically improving
    convergence speed in some contexts.

    Original Paper:
    Sun, Yanxia & Qi, Guoyuan & Wang, Zenghui & Van Wyk, Barend & Hamam, Yskandar. (2009). Chaotic particle
    swarm optimization. 505-510. 10.1145/1543834.1543902.

    https://www.researchgate.net/publication/220741402_Chaotic_particle_swarm_optimization
    """
    def __init__(self,
                 params: Iterable[torch.nn.Parameter],
                 num_particles: int = 100,
                 a: float = 0.02,
                 b: float = 0.01,
                 c: float = 0.01,
                 beta: float = .001,
                 k: float = 15.,
                 epsilon: float = 1.,
                 i0: float = 0.2,
                 z: float = 0.7,
                 max_param_value: float = -10,
                 min_param_value: float = 10):
        particle_kwargs = {'a': a,
                           'b': b,
                           'c': c,
                           'beta': beta,
                           'k': k,
                           'epsilon': epsilon,
                           'i0': i0,
                           'z':z,
                           'max_param_value': max_param_value,
                           'min_param_value': min_param_value,
                           }
        super().__init__(params, num_particles, particle_class=ChaoticParticle, particle_kwargs=particle_kwargs)



from .optim.ParticleSwarmOptimizer import ParticleSwarmOptimizer
from .optim.GenerationalPSO import GenerationalPSO
from .optim.AutotuningPSO import AutotuningPSO
from .optim.RingTopologyPSO import RingTopologyPSO
from .optim.ChaoticPSO import ChaoticPSO
from .optim.GenericPSO import GenericPSO
from .optim.AcceleratedPSO import AcceleratedPSO
from .optim.SineCosineAlgorithm import SineCosineAlgorithm
from .optim.ImprovedSineCosineAlgorithm import ImprovedSineCosineAlgorithm

OPTIMIZERS = list(GenericPSO.subclasses)


from typing import List

import pytest
import torch
from torch import Tensor

from torch_pso import OPTIMIZERS


class SquareWeightsModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.weights = torch.nn.Parameter(torch.rand((1,)))

    def forward(self, x):
        x = self.weights
        return x ** 2


class SquarePlus2WeightsModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.weights = torch.nn.Parameter(torch.rand((1,)))

    def forward(self, x):
        x = self.weights
        return (x + 2) ** 2


class QuarticWeightsModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.weights = torch.nn.Parameter(torch.rand((1,)))

    def forward(self, x):
        x = self.weights
        return x ** 2 * (x ** 2 - 1)


class RastriginModule(torch.nn.Module):
    def __init__(self, num_dimensions, a=10.):
        super().__init__()
        self.weights = torch.nn.Parameter(torch.rand((num_dimensions,)))
        self.num_dimensions = num_dimensions
        self.A = a

    def forward(self, x):
        x = self.weights
        return self.A * self.num_dimensions + (x ** 2 - self.A * torch.cos(2 * torch.pi * x)).sum()


class HimmelblauModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.weights = torch.nn.Parameter(torch.rand((2,)))

    def forward(self, x):
        x, y = self.weights
        return (x ** 2 + y - 11) ** 2 + (x + y ** 2 - 7) ** 2


class GoldsteinPriceModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.weights = torch.nn.Parameter(torch.rand((2,)))

    def forward(self, x):
        x, y = self.weights
        return (1 + (x+y+1)**2*(19-14*x+3*x**2-14*y+6*x*y+3*y**2))*(30 + (2*x-3*y)**2*(18-32*x+12*x**2+48*y-36*x*y+27*y**2))


@pytest.mark.parametrize('optimizer_type', OPTIMIZERS)
def test_square_converges(optimizer_type):
    if optimizer_type.__name__ == 'RingTopologyPSO':
        # Ring Topology PSO converges SUPER slowly on a simple quadratic function.
        # Sufficiently so that even passing this test is hit or miss
        return
    net = SquareWeightsModule()
    optim = optimizer_type(net.parameters())

    def closure():
        optim.zero_grad()
        return net(None)

    # Make sure we don't start out satisfying the condition
    assert not torch.allclose(net.weights, torch.Tensor([0.]), atol=1e-4, rtol=1e-3)

    converged = False
    for _ in range(10000):
        optim.step(closure)
        if torch.allclose(net.weights, torch.Tensor([0.]), atol=1e-2, rtol=1e-1):
            converged = True
            break

    assert converged, net.weights


@pytest.mark.parametrize('optimizer_type', OPTIMIZERS)
def test_square_plus_2_converges(optimizer_type):
    if optimizer_type.__name__ == 'RingTopologyPSO':
        # Ring Topology PSO converges SUPER slowly on a simple quadratic function.
        # Sufficiently so that even passing this test is hit or miss
        return
    net = SquarePlus2WeightsModule()
    optim = optimizer_type(net.parameters())

    def closure():
        optim.zero_grad()
        return net(None)

    # Make sure we don't start out satisfying the condition
    assert not torch.allclose(net.weights, torch.Tensor([-2.]), atol=1e-4, rtol=1e-3)

    converged = False
    for _ in range(1000):
        optim.step(closure)
        if torch.allclose(net.weights, torch.Tensor([-2.]), atol=1e-2, rtol=1e-1):
            converged = True
            break

    assert converged, net.weights


@pytest.mark.parametrize('optimizer_type', OPTIMIZERS)
def test_quartic_converges(optimizer_type):
    if optimizer_type.__name__ == 'RingTopologyPSO':
        # Ring Topology PSO converges SUPER slowly on a simple quadratic function.
        # Sufficiently so that even passing this test is hit or miss
        return
    net = QuarticWeightsModule()
    optim = optimizer_type(net.parameters())

    def closure():
        optim.zero_grad()
        return net(None)

    # Make sure we don't start out satisfying the condition
    assert not torch.allclose(abs(net.weights), torch.sqrt(torch.Tensor([2])) / 2, atol=1e-4, rtol=1e-3)

    converged = False
    for _ in range(10000):
        optim.step(closure)
        if torch.allclose(abs(net.weights), torch.sqrt(torch.Tensor([2])) / 2, atol=1e-2, rtol=1e-1):
            converged = True
            break

    assert converged, net.weights


@pytest.mark.parametrize('optimizer_type', OPTIMIZERS)
def test_rastrigin_converges(optimizer_type):
    if optimizer_type.__name__ == 'RingTopologyPSO':
        # Ring Topology PSO converges SUPER slowly on a simple quadratic function.
        # Sufficiently so that even passing this test is hit or miss
        return
    net = RastriginModule(num_dimensions=1)
    optim = optimizer_type(net.parameters())

    global_minimum = torch.zeros_like(net.weights)

    def closure():
        optim.zero_grad()
        return net(None)

    # Make sure we don't start out satisfying the condition
    assert not torch.allclose(net.weights, global_minimum, atol=1e-4, rtol=1e-3)

    converged = False
    for _ in range(10000):
        optim.step(closure)
        if torch.allclose(net.weights, global_minimum, atol=1e-2, rtol=1e-1):
            converged = True
            break

    assert converged, net.weights


@pytest.mark.parametrize('optimizer_type', OPTIMIZERS)
def test_rastrigin3_converges(optimizer_type):
    if optimizer_type.__name__ in ['RingTopologyPSO', 'GenerationalPSO', 'ChaoticPSO', 'AcceleratedPSO']:
        # These PSO algorithms converge very slowly on this problem
        return
    net = RastriginModule(num_dimensions=3)
    optim = optimizer_type(net.parameters())

    global_minimum = torch.zeros_like(net.weights)

    def closure():
        optim.zero_grad()
        return net(None)

    # Make sure we don't start out satisfying the condition
    assert not torch.allclose(net.weights, global_minimum, atol=1e-4, rtol=1e-3)

    converged = False
    for _ in range(5000):
        optim.step(closure)
        if torch.allclose(net.weights, global_minimum, atol=1e-2, rtol=1e-1):
            converged = True
            break

    assert converged, net.weights


# All of the algorithms converge really slowly on this test. Interestingly, there is always a single dimension
# That converges more slowly than the rest.
@pytest.mark.skip('Difficult convergence test.')
@pytest.mark.parametrize('optimizer_type', OPTIMIZERS)
def test_rastrigin10_converges(optimizer_type):
    if optimizer_type.__name__ in ['RingTopologyPSO']:
        # These PSO algorithms converge very slowly on this problem
        return
    net = RastriginModule(num_dimensions=10)
    optim = optimizer_type(net.parameters())

    global_minimum = torch.zeros_like(net.weights)

    def closure():
        optim.zero_grad()
        return net(None)

    # Make sure we don't start out satisfying the condition
    assert not torch.allclose(net.weights, global_minimum, atol=1e-4, rtol=1e-3)

    converged = False
    for _ in range(10000):
        optim.step(closure)
        if torch.allclose(net.weights, global_minimum, atol=1e-2, rtol=1e-1):
            converged = True
            break

    assert converged, net.weights


@pytest.mark.parametrize('optimizer_type', OPTIMIZERS)
def test_himmelblau_converges(optimizer_type):
    if optimizer_type.__name__ in ['RingTopologyPSO']:
        # These PSO algorithms converge very slowly on this problem
        return
    net = HimmelblauModule()
    optim = optimizer_type(net.parameters())

    global_minima = [torch.Tensor([3., 2.]),
                     torch.Tensor([-2.805118, 3.131312]),
                     torch.Tensor([-3.779310, -3.283186]),
                     torch.Tensor([3.584428, -1.848126]),
                     ]

    def close_to_a_minimum(x, minima: List[Tensor]) -> bool:
        return any(torch.allclose(x, m, atol=1e-2, rtol=1e-1) for m in minima)

    def closure():
        optim.zero_grad()
        return net(None)

    # Make sure we don't start out satisfying the condition
    assert not close_to_a_minimum(net.weights, global_minima)

    converged = False
    for _ in range(4000):
        optim.step(closure)
        if close_to_a_minimum(net.weights, global_minima):
            converged = True
            break

    assert converged, net.weights


@pytest.mark.parametrize('optimizer_type', OPTIMIZERS)
def test_goldstein_price_converges(optimizer_type):
    if optimizer_type.__name__ in ['RingTopologyPSO', 'SineCosineAlgorithm', 'ImprovedSineCosineAlgorithm']:
        # These PSO algorithms converge very slowly on this problem
        return
    net = GoldsteinPriceModule()
    optim = optimizer_type(net.parameters())

    global_minima = [torch.Tensor([0, -1]),
                     ]

    def close_to_a_minimum(x, minima: List[Tensor]) -> bool:
        return any(torch.allclose(x, m, atol=1e-1, rtol=1e-1) for m in minima)

    def closure():
        optim.zero_grad()
        return net(None)

    # Make sure we don't start out satisfying the condition
    assert not close_to_a_minimum(net.weights, global_minima)

    converged = False
    for _ in range(4000):
        optim.step(closure)
        if close_to_a_minimum(net.weights, global_minima):
            converged = True
            break

    assert converged, net.weights



from typing import Callable, List, Dict, Iterable

import torch

from .GenericPSO import clone_param_groups, _initialize_param_groups, GenericParticle, GenericPSO


class AcceleratedParticle(GenericParticle):
    def __init__(self,
                 param_groups,
                 alpha: float,
                 beta: float,
                 decay_parameter:float,
                 max_param_value: float,
                 min_param_value: float):
        if not 0 < beta < 1:
            raise ValueError(f'Beta must be a positive constant, not of value {beta}.')
        if not 0 < decay_parameter < 1:
            raise ValueError(f'Decay parameter must be a positive constant, not of value {decay_parameter}.')

        self.param_groups = param_groups
        self.alpha = alpha
        self.beta = beta
        self.decay_parameter = decay_parameter
        self.max_param_value = max_param_value
        self.min_param_value = min_param_value

        self.position = _initialize_param_groups(param_groups, max_param_value, min_param_value)

        self.best_known_position = clone_param_groups(self.position)
        self.best_known_loss_value = torch.inf

    def step(self, closure: Callable[[], torch.Tensor], global_best_param_groups: List[Dict]) -> torch.Tensor:
        """
        Particle will take one step.
        :param closure: A callable that reevaluates the model and returns the loss.
        :param global_best_param_groups: List of param_groups that yield the best found loss globally
        :return:
        """
        # Because our parameters are not a single tensor, we have to iterate over each group, and then each param in
        # each group.
        for position_group, global_best, master in zip(self.position,
                                                       global_best_param_groups,
                                                       self.param_groups):
            position_group_params = position_group['params']
            global_best_params = global_best['params']
            master_params = master['params']

            new_position_params = []
            for p, gb, m in zip(position_group_params,
                                global_best_params, master_params):
                rand_group = torch.rand_like(p)
                new_position = (1-self.beta)*p + self.beta*gb + self.alpha*rand_group
                new_position_params.append(new_position)
                m.data = new_position.data  # Update the model, so we can use it for calculating loss
            position_group['params'] = new_position_params

        # Really crummy way to update the parameter weights in the original model.
        # Simply changing self.param_groups doesn't update the model.
        # Nor does changing its elements or the raw values of 'param' of the elements.
        # We have to change the underlying tensor data to point to the new positions
        for i in range(len(self.position)):
            for j in range(len(self.param_groups[i]['params'])):
                self.param_groups[i]['params'][j].data = self.param_groups[i]['params'][j].data

        # Calculate new loss after moving and update the best known position if we're in a better spot
        new_loss = closure()
        if new_loss < self.best_known_loss_value:
            self.best_known_position = clone_param_groups(self.position)
            self.best_known_loss_value = new_loss
        return new_loss


class AcceleratedPSO(GenericPSO):
    """
    ChaoticPSO is a variation on the original Particle Swarm Optimization developed to quickly train SVMs. It simplifies
    the original PSO algorithm--most notably eschewing the need for velocity states.

    Original Paper:
    Yang, X. S., Deb, S., and Fong, S., (2011), Accelerated Particle Swarm Optimization and Support Vector Machine for
    Business Optimization and Applications, in: NDT2011, CCIS 136, Springer, pp. 53-66 (2011)

    https://arxiv.org/abs/1203.6577
    """

    def __init__(self,
                 params: Iterable[torch.nn.Parameter],
                 num_particles: int = 100,
                 alpha: float = 0.3,
                 beta: float = 0.45,
                 decay_parameter: float = 0.7,
                 max_param_value: float = -10,
                 min_param_value: float = 10):
        particle_kwargs = {'alpha': alpha,
                           'beta': beta,
                           'decay_parameter': decay_parameter,
                           'max_param_value': max_param_value,
                           'min_param_value': min_param_value,
                           }
        super().__init__(params, num_particles, particle_class=AcceleratedParticle, particle_kwargs=particle_kwargs)
